"""
miniwdl-aws-submit CLI entry point (console script) to submit a miniwdl "workflow job" to an AWS
Batch queue, which will invoke miniwdl-run-s3upload to run the workflow (spawning additional Batch
jobs as needed to execute tasks). This is typically used on-laptop to kick off workflows, without
the laptop needing to stay on/connected. It can also wait for the workflow job to complete and
stream its logs.
"""

import sys
import os
import time
import argparse
import shlex
from datetime import datetime
from collections import defaultdict
import boto3
from ._util import detect_aws_region, randomize_job_name, END_OF_LOG, efs_id_from_access_point


def miniwdl_submit_awsbatch(argv):
    # Configure from arguments/environment/tags
    args, unused_args = parse_args(argv)
    verbose = args.follow or "--verbose" in unused_args or "--debug" in unused_args
    detect_env_args(args)
    if verbose:
        print("Workflow job queue: " + args.workflow_queue, file=sys.stderr)

    aws_region_name = detect_aws_region(None)
    if not aws_region_name:
        print(
            "Failed to detect AWS region; configure AWS CLI or set environment AWS_DEFAULT_REGION",
            file=sys.stderr,
        )
        sys.exit(1)
    aws_batch = boto3.client("batch", region_name=aws_region_name)
    detect_tags_args(aws_batch, args)

    if verbose:
        print("Workflow IAM role ARN: " + args.workflow_role, file=sys.stderr)
        print("Task job queue: " + args.task_queue, file=sys.stderr)
        print("EFS Access Point: " + args.fsap, file=sys.stderr)

    fs_id = efs_id_from_access_point(aws_region_name, args.fsap)
    if verbose:
        print("EFS: " + fs_id, file=sys.stderr)

    # TODO: accept local WDL source code; check, zip, & attach it

    # Prepare workflow job: command, environment, and container properties
    job_name, miniwdl_run_cmd = form_miniwdl_run_cmd(args, unused_args)
    job_name = randomize_job_name(job_name)

    environment = [
        {"name": "MINIWDL__AWS__FS", "value": fs_id},
        {"name": "MINIWDL__AWS__FSAP", "value": args.fsap},
        {"name": "MINIWDL__AWS__TASK_QUEUE", "value": args.task_queue},
        {"name": "MINIWDL__FILE_IO__ROOT", "value": args.mount},
    ]
    extra_env = set()
    if not args.no_env:
        # pass through environment variables starting with MINIWDL__ (except those specific to
        # workflow job launch, or passed through via command line)
        for k in os.environ:
            if k.startswith("MINIWDL__") and k not in (
                "MINIWDL__AWS__FS",
                "MINIWDL__AWS__FSAP",
                "MINIWDL__AWS__TASK_QUEUE",
                "MINIWDL__AWS__WORKFLOW_QUEUE",
                "MINIWDL__AWS__WORKFLOW_ROLE",
                "MINIWDL__AWS__WORKFLOW_IMAGE",
                "MINIWDL__AWS__S3_UPLOAD_FOLDER",
                "MINIWDL__AWS__S3_UPLOAD_DELETE_AFTER",
                "MINIWDL__FILE_IO__ROOT",
            ):
                environment.append({"name": k, "value": os.environ[k]})
                extra_env.add(k)

    if verbose:
        print("Workflow job image: " + args.image, file=sys.stderr)
        if extra_env:
            print(
                "Passing through environment variables (--no-env to disable): "
                + " ".join(list(extra_env)),
                file=sys.stderr,
            )
        print("Invocation: " + " ".join(shlex.quote(s) for s in miniwdl_run_cmd), file=sys.stderr)

    workflow_container_props = {
        "fargatePlatformConfiguration": {"platformVersion": "1.4.0"},
        "executionRoleArn": args.workflow_role,
        "jobRoleArn": args.workflow_role,
        "resourceRequirements": [
            {"type": "VCPU", "value": str(args.cpu)},
            {"type": "MEMORY", "value": str(args.memory_GiB * 1024)},
        ],
        "volumes": [
            {
                "name": "efs",
                "efsVolumeConfiguration": {
                    "fileSystemId": fs_id,
                    "transitEncryption": "ENABLED",
                    "authorizationConfig": {"accessPointId": args.fsap},
                },
            }
        ],
        "mountPoints": [{"containerPath": args.mount, "sourceVolume": "efs"}],
        "image": args.image,
        "command": miniwdl_run_cmd,
        "environment": environment,
    }
    if not args.no_public_ip:
        workflow_container_props["networkConfiguration"] = {"assignPublicIp": "ENABLED"}

    # Register & submit workflow job
    workflow_job_def = aws_batch.register_job_definition(
        jobDefinitionName=job_name,
        platformCapabilities=["FARGATE"],
        type="container",
        containerProperties=workflow_container_props,
    )
    workflow_job_def_handle = (
        f"{workflow_job_def['jobDefinitionName']}:{workflow_job_def['revision']}"
    )
    try:
        workflow_job_id = aws_batch.submit_job(
            jobName=job_name,
            jobQueue=args.workflow_queue,
            jobDefinition=workflow_job_def_handle,
        )["jobId"]
        if verbose:
            print(f"Submitted {job_name} to {args.workflow_queue}:", file=sys.stderr)
            sys.stderr.flush()
        print(workflow_job_id)
        if not sys.stdout.isatty():
            print(workflow_job_id, file=sys.stderr)
    finally:
        aws_batch.deregister_job_definition(jobDefinition=workflow_job_def_handle)

    # Wait for workflow job, if requested
    exit_code = 0
    if args.wait or args.follow:
        exit_code = wait(
            aws_region_name,
            aws_batch,
            workflow_job_id,
            args.follow,
            expect_log_eof=not args.self_test,
        )
    sys.exit(exit_code)


def parse_args(argv):
    if "COLUMNS" not in os.environ:
        os.environ["COLUMNS"] = "100"
    parser = argparse.ArgumentParser(
        prog="miniwdl-aws-submit",
        description="Launch `miniwdl run` on AWS Batch (+ EFS at /mnt/efs), itself launching additional"
        " Batch jobs to execute WDL tasks. Passed-through arguments to `miniwdl run` should refer to"
        " s3:// or /mnt/efs/ input paths, rather than the local filesystem.",
        usage="miniwdl-aws-submit [miniwdl_run_arg ...] --workflow-queue WORKFLOW_QUEUE",
        allow_abbrev=False,
    )
    group = parser.add_argument_group("AWS Batch")
    group.add_argument(
        "--workflow-queue",
        help="job queue for workflow job [env MINIWDL__AWS__WORKFLOW_QUEUE]",
    )
    group.add_argument(
        "--task-queue",
        help="job queue for task jobs [env MINIWDL__AWS__TASK_QUEUE"
        " or detect from DefaultTaskQueue tag on workflow job queue",
    )
    group.add_argument(
        "--fsap",
        help="EFS Access Point ID (fsap-xxxx) for mounting [env MINIWDL__AWS__FSAP"
        " or detect from DefaultFsap tag on workflow job queue]",
    )
    group.add_argument(
        "--mount", default="/mnt/efs", help="EFS mount point in all containers [/mnt/efs]"
    )
    group = parser.add_argument_group("Workflow job provisioning")
    group.add_argument(
        "--workflow-role",
        help="ARN of execution+job role for workflow job [env MINIWDL__AWS__WORKFLOW_ROLE"
        " or detect from WorkflowEngineRoleArn tag on workflow job queue]",
    )
    group.add_argument("--name", help="workflow job name [WDL filename]")
    group.add_argument(
        "--cpu", metavar="N", type=str, default="1", help="vCPUs for workflow job [1]"
    )
    group.add_argument(
        "--memory-GiB", metavar="N", type=int, default=4, help="memory for workflow job [4]"
    )
    group.add_argument(
        "--image",
        help="override miniwdl-aws Docker image tag for workflow job [env MINIWDL__AWS__WORKFLOW_IMAGE]",
    )
    group.add_argument(
        "--no-env", action="store_true", help="don't pass through MINIWDL__* environment variables"
    )
    group.add_argument(
        "--no-public-ip",
        action="store_true",
        help="don't assign public IP (workflow compute env has private subnet & NAT)",
    )
    group = parser.add_argument_group("miniwdl I/O")
    group.add_argument(
        "--dir",
        default=None,
        help="Run directory prefix [/mnt/efs/miniwdl_run]",
    )
    group.add_argument(
        "--s3upload",
        help="s3://bucket/folder/ at which to upload run outputs (otherwise left on EFS)",
    )
    group.add_argument(
        "--delete-after",
        choices=("always", "success", "failure"),
        help="with --s3upload, delete EFS run directory afterwards",
    )
    parser.add_argument(
        "--wait", "-w", action="store_true", help="wait for workflow job to complete"
    )
    parser.add_argument(
        "--follow",
        "-f",
        action="store_true",
        help="live-stream workflow log to standard error (implies --wait)",
    )
    parser.add_argument("--self-test", action="store_true", help="perform `miniwdl run_self_test`")

    args, unused_args = parser.parse_known_args(argv[1:])

    if args.mount.endswith("/"):
        args.mount = args.mount[:-1]
    assert args.mount
    if not args.dir:
        args.dir = os.path.join(args.mount, "miniwdl_run")
    if not args.dir.startswith(args.mount):
        print(f"--dir must begin with {args.mount}", file=sys.stderr)
        sys.exit(1)

    return (args, unused_args)


def detect_env_args(args):
    """
    Detect configuration set through environment variables (that weren't set by command-line args)
    """
    args.fsap = args.fsap if args.fsap else os.environ.get("MINIWDL__AWS__FSAP", "")
    args.workflow_queue = (
        args.workflow_queue
        if args.workflow_queue
        else os.environ.get("MINIWDL__AWS__WORKFLOW_QUEUE", None)
    )
    if not args.workflow_queue:
        print(
            "--workflow-queue is required (or environment variable MINIWDL__AWS__WORKFLOW_QUEUE)",
            file=sys.stderr,
        )
        sys.exit(1)
    args.fsap = args.fsap if args.fsap else os.environ.get("MINIWDL__AWS__FSAP", "")
    args.task_queue = (
        args.task_queue if args.task_queue else os.environ.get("MINIWDL__AWS__TASK_QUEUE", None)
    )
    args.workflow_role = (
        args.workflow_role
        if args.workflow_role
        else os.environ.get("MINIWDL__AWS__WORKFLOW_ROLE", None)
    )
    args.image = args.image if args.image else os.environ.get("MINIWDL__AWS__WORKFLOW_IMAGE", None)
    if not args.image:
        import importlib_metadata

        try:
            args.image = "ghcr.io/miniwdl-ext/miniwdl-aws:v" + importlib_metadata.version(
                "miniwdl-aws"
            )
        except importlib_metadata.PackageNotFoundError:
            print(
                "Failed to detect miniwdl Docker image version tag; set explicitly with --image or MINIWDL__AWS__WORKFLOW_IMAGE",
                file=sys.stderr,
            )
            sys.exit(1)
    if args.delete_after and not args.s3upload:
        print("--delete-after requires --s3upload", file=sys.stderr)
        sys.exit(1)
    args.s3upload = (
        args.s3upload if args.s3upload else os.environ.get("MINIWDL__AWS__S3_UPLOAD_FOLDER", None)
    )
    args.delete_after = (
        args.delete_after.strip().lower()
        if args.delete_after
        else os.environ.get("MINIWDL__AWS__DELETE_AFTER_S3_UPLOAD", None)
    )


def detect_tags_args(aws_batch, args):
    """
    If not otherwise set by command line arguments or environment, inspect tags of the workflow job
    queue to detect default EFS Access Point ID (fsap-xxx), task job queue, and/or workflow role
    ARN. Infra provisioning (CloudFormation, Terraform, etc.) may have set the respective tags.
    """
    if not (args.fsap or args.task_queue or args.workflow_role):
        workflow_queue_tags = aws_batch.describe_job_queues(jobQueues=[args.workflow_queue])[
            "jobQueues"
        ][0]["tags"]
        if not args.fsap:
            try:
                args.fsap = workflow_queue_tags["DefaultFsap"]
                assert args.fsap.startswith("fsap-")
            except:
                print(
                    "Unable to detect default EFS Access Point (fsap-xxxx) from DefaultFsap tag of workflow job queue."
                    " Set --fsap or environment variable MINIWDL__AWS__FSAP.",
                    file=sys.stderr,
                )
                sys.exit(1)
        if not args.task_queue:
            args.task_queue = workflow_queue_tags.get("DefaultTaskQueue", None)
            if not args.task_queue:
                print(
                    "Unable to detect default task job queue name from DefaultTaskQueue tag of workflow job queue."
                    " Set --task-queue or environment variable MINIWDL__AWS__TASK_QUEUE.",
                    file=sys.stderr,
                )
                sys.exit(1)
        if not args.workflow_role:
            # Workflow role ARN is needed for Fargate Batch (unlike EC2 Batch, where a role is
            # associated with the EC2 instance profile in the compute environment).
            try:
                args.workflow_role = aws_batch.describe_job_queues(jobQueues=[args.workflow_queue])[
                    "jobQueues"
                ][0]["tags"]["WorkflowEngineRoleArn"]
                assert args.workflow_role.startswith("arn:aws:iam::")
            except:
                if not args.workflow_role:
                    print(
                        "Unable to detect ARN of workflow engine IAM role from WorkflowEngineRoleArn tag of workflow job queue."
                        " Double-check --workflow-queue, or set --workflow-role or environment MINIWDL__AWS__WORKFLOW_ROLE.",
                        file=sys.stderr,
                    )
                    sys.exit(1)


def form_miniwdl_run_cmd(args, unused_args):
    """
    Formulate the `miniwdl run` command line to be invoked in the workflow job container
    """
    if args.self_test:
        self_test_dir = os.path.join(
            args.mount, "miniwdl_run_self_test", datetime.today().strftime("%Y%m%d_%H%M%S")
        )
        miniwdl_run_cmd = ["miniwdl", "run_self_test", "--dir", self_test_dir]
        job_name = args.name if args.name else "miniwdl_run_self_test"
    else:
        wdl_filename = next((arg for arg in unused_args if not arg.startswith("-")), None)
        if not wdl_filename:
            print("Command line appears to be missing WDL filename", file=sys.stderr)
            sys.exit(1)
        job_name = args.name
        if not job_name:
            job_name = os.path.basename(wdl_filename).lstrip(".")
            try:
                for punct in (".", "?"):
                    if job_name.index(punct) > 0:
                        job_name = job_name[: job_name.index(punct)]
            except ValueError:
                pass
            job_name = ("miniwdl_run_" + job_name)[:128]
        # pass most arguments through to miniwdl-run-s3upload inside workflow job
        miniwdl_run_cmd = ["miniwdl-run-s3upload"] + unused_args
        miniwdl_run_cmd.extend(["--dir", args.dir])
        miniwdl_run_cmd.extend(["--s3upload", args.s3upload] if args.s3upload else [])
        miniwdl_run_cmd.extend(["--delete-after", args.delete_after] if args.delete_after else [])
    return (job_name, miniwdl_run_cmd)


def wait(aws_region_name, aws_batch, workflow_job_id, follow, expect_log_eof=True):
    """
    Wait for workflow job to complete & return its exit code; optionally tail its log to stderr
    """
    try:
        log_follower = None
        exit_code = None
        saw_end = False
        while exit_code is None:
            time.sleep(1.0)
            job_descs = aws_batch.describe_jobs(jobs=[workflow_job_id])
            job_desc = job_descs["jobs"][0]
            if (
                not log_follower
                and "container" in job_desc
                and "logStreamName" in job_desc["container"]
            ):
                log_stream_name = job_desc["container"]["logStreamName"]
                print("Log stream: " + log_stream_name, file=sys.stderr)
                sys.stderr.flush()
                log_follower = CloudWatchLogsFollower(
                    boto3.DEFAULT_SESSION, aws_region_name, "/aws/batch/job", log_stream_name
                )
            if follow and log_follower:
                for event in log_follower.new_events():
                    if END_OF_LOG not in event["message"]:
                        print(event["message"], file=sys.stderr)
                    else:
                        saw_end = True
                sys.stderr.flush()
            if job_desc["status"] == "SUCCEEDED":
                exit_code = 0
            elif job_desc["status"] == "FAILED":
                exit_code = -1
                if "container" in job_desc and "exitCode" in job_desc["container"]:
                    exit_code = job_desc["container"]["exitCode"]
                    assert exit_code != 0
        if expect_log_eof and follow and log_follower and not saw_end:
            # give straggler log messages a few seconds to appear
            time.sleep(3.0)
            for event in log_follower.new_events():
                if END_OF_LOG not in event["message"]:
                    print(event["message"], file=sys.stderr)
                else:
                    saw_end = True
            if not saw_end:
                print(
                    f"[miniwdl-aws-submit] WARNING: end-of-log marker not seen; more information may appear in log stream {log_stream_name}",
                    file=sys.stderr,
                )
            sys.stderr.flush()
        print(job_desc["status"] + "\t" + workflow_job_id, file=sys.stderr)
        assert isinstance(exit_code, int)
        return exit_code
    except KeyboardInterrupt:
        print(
            f"[miniwdl-aws-submit] interrupted by Ctrl-C; workflow may remain active in workflow job {workflow_job_id}",
            file=sys.stderr,
        )
        return -1


class CloudWatchLogsFollower:
    # Based loosely on:
    #   https://github.com/aws/aws-cli/blob/v2/awscli/customizations/logs/tail.py
    # which wasn't suitable to use directly at the time of this writing, because of
    #   https://github.com/aws/aws-cli/issues/5560
    def __init__(self, boto_session, region_name, group_name, stream_name=None):
        self.group_name = group_name
        self.stream_name = stream_name
        self._newest_timestamp = None
        self._newest_event_ids = set()
        self._client = boto_session.client("logs", region_name=region_name)

    def new_events(self):
        event_ids_per_timestamp = defaultdict(set)

        filter_args = {"logGroupName": self.group_name}
        if self.stream_name:
            filter_args["logStreamNames"] = [self.stream_name]
        if self._newest_timestamp:
            filter_args["startTime"] = self._newest_timestamp
        while True:
            try:
                response = self._client.filter_log_events(**filter_args)
            except self._client.exceptions.ResourceNotFoundException:
                return  # we may learn the Batch job's log stream name before it actually exists
            for event in response["events"]:
                # For the case where we've hit the last page, we will be
                # reusing the newest timestamp of the received events to keep polling.
                # This means it is possible that duplicate log events with same timestamp
                # are returned back which we do not want to yield again.
                # We only want to yield log events that we have not seen.
                if event["eventId"] not in self._newest_event_ids:
                    event_ids_per_timestamp[event["timestamp"]].add(event["eventId"])
                    yield event
            if "nextToken" in response:
                filter_args["nextToken"] = response["nextToken"]
            else:
                break

        if event_ids_per_timestamp:
            self._newest_timestamp = max(event_ids_per_timestamp.keys())
            self._newest_event_ids = event_ids_per_timestamp[self._newest_timestamp]



"""
BatchJob: implements miniwdl TaskContainer by submitting jobs to an AWS Batch queue and polling
their status. Assumes a shared filesystem (typically EFS) between the miniwdl host and the Batch
workers.
"""

import os
import math
import time
import threading
import heapq
from contextlib import ExitStack, suppress
import boto3
import botocore
import WDL
import WDL.runtime.task_container
import WDL.runtime._statusbar
from WDL._util import PygtailLogger, rmtree_atomic, symlink_force
from WDL._util import StructuredLogMessage as _
from ._util import (
    detect_aws_region,
    randomize_job_name,
    efs_id_from_access_point,
    detect_sagemaker_studio_efs,
    detect_studio_fsap,
    detect_gwfcore_batch_queue,
)


class BatchJob(WDL.runtime.task_container.TaskContainer):
    @classmethod
    def global_init(cls, cfg, logger):
        cls._set_config_defaults(cfg, logger)
        cls._region_name = detect_aws_region(cfg)
        assert (
            cls._region_name
        ), "Failed to detect AWS region; configure AWS CLI or set environment AWS_DEFAULT_REGION"

        # EFS configuration based on:
        # - [aws] fsap / MINIWDL__AWS__FSAP
        # - [aws] fs / MINIWDL__AWS__FS
        # - SageMaker Studio metadata, if applicable
        cls._fs_id = None
        cls._fsap_id = None
        cls._fs_mount = cfg.get("file_io", "root")
        assert (
            len(cls._fs_mount) > 1
        ), "misconfiguration, set [file_io] root / MINIWDL__FILE_IO__ROOT to EFS mount point"
        if cfg.has_option("aws", "fs"):
            cls._fs_id = cfg.get("aws", "fs")
        if cfg.has_option("aws", "fsap"):
            cls._fsap_id = cfg.get("aws", "fsap")
            if not cls._fs_id:
                cls._fs_id = efs_id_from_access_point(cls._region_name, cls._fsap_id)
        cls._studio_efs_uid = None
        sagemaker_studio_efs = detect_sagemaker_studio_efs(logger, region_name=cls._region_name)
        if sagemaker_studio_efs:
            (
                studio_efs_id,
                studio_efs_uid,
                studio_efs_home,
                studio_efs_mount,
            ) = sagemaker_studio_efs
            assert (
                not cls._fs_id or cls._fs_id == studio_efs_id
            ), "Configured EFS ([aws] fs / MINIWDL__AWS__FS, [aws] fsap / MINIWDL__AWS__FSAP) isn't associated with current SageMaker Studio domain EFS"
            cls._fs_id = studio_efs_id
            assert cls._fs_mount.rstrip("/") == studio_efs_mount.rstrip("/"), (
                "misconfiguration, set [file_io] root / MINIWDL__FILE_IO__ROOT to "
                + studio_efs_mount.rstrip("/")
            )
            cls._studio_efs_uid = studio_efs_uid
            if not cls._fsap_id:
                cls._fsap_id = detect_studio_fsap(
                    logger,
                    studio_efs_id,
                    studio_efs_uid,
                    studio_efs_home,
                    region_name=cls._region_name,
                )
                assert (
                    cls._fsap_id
                ), "Unable to detect suitable EFS Access Point for use with SageMaker Studio; set [aws] fsap / MINIWDL__AWS__FSAP"
            # TODO: else sanity-check that FSAP's root directory equals studio_efs_home
        assert (
            cls._fs_id
        ), "Missing EFS configuration ([aws] fs / MINIWDL__AWS__FS or [aws] fsap / MINIWDL__AWS__FSAP)"
        if not cls._fsap_id:
            logger.warning(
                "AWS BatchJob plugin recommends using EFS Access Point to simplify permissions between containers (configure [aws] fsap / MINIWDL__AWS__FSAP to fsap-xxxx)"
            )
        logger.debug(
            _(
                "AWS BatchJob EFS configuration",
                fs_id=cls._fs_id,
                fsap_id=cls._fsap_id,
                mount=cls._fs_mount,
            )
        )

        # set AWS Batch job queue
        if cfg.has_option("aws", "task_queue"):
            cls._job_queue = cfg.get("aws", "task_queue")
        elif sagemaker_studio_efs:
            cls._job_queue = detect_gwfcore_batch_queue(
                logger, sagemaker_studio_efs[0], region_name=cls._region_name
            )
        assert (
            cls._job_queue
        ), "Missing AWS Batch job queue configuration ([aws] task_queue / MINIWDL__AWS__TASK_QUEUE)"

        # TODO: query Batch compute environment for resource limits
        cls._resource_limits = {"cpu": 9999, "mem_bytes": 999999999999999}
        cls._submit_lock = threading.Lock()
        cls._last_submit_time = [0.0]
        cls._init_time = time.time()
        cls._describer = BatchJobDescriber()
        logger.info(
            _(
                "initialized AWS BatchJob plugin",
                region_name=cls._region_name,
                job_queue=cls._job_queue,
                resource_limits=cls._resource_limits,
            )
        )

    @classmethod
    def _set_config_defaults(cls, cfg, logger):
        # Set defaults in the cfg: ConfigLoader object so that our subsequent cfg.get* ops won't
        # throw if the user's custom .cfg file, which we don't necessarily control, omits them.
        #
        # This repo's miniwdl_aws.cfg also sets these options, and takes precedence in typical
        # operations using the auto-built Docker image. But consider the case where a user has
        # edited/written their own .cfg file, and we release a new plugin version with a new
        # config option. That'd break their existing .cfg file unless the option has a default
        # of last resort set here.
        cfg.plugin_defaults(
            {
                "aws": {
                    "job_timeout": 864000,
                    "describe_period": 1,
                    "submit_period": 1,
                    "submit_period_b": 0.0,
                    "submit_period_c": 0.0,
                    "boto3_retries": {"max_attempts": 5, "mode": "standard"},
                    "retry_wait": 20,
                    "container_sync": False,
                    "job_tags": dict(),
                }
            }
        )

    @classmethod
    def detect_resource_limits(cls, cfg, logger):
        return cls._resource_limits

    def __init__(self, cfg, run_id, host_dir):
        super().__init__(cfg, run_id, host_dir)
        self._logStreamName = None
        self._inputs_copied = False
        # We'll direct Batch to mount EFS inside the task container at the same location we have
        # it mounted ourselves, namely /mnt/efs. Therefore container_dir will be the same as
        # host_dir (unlike the default Swarm backend, which mounts it at a different virtualized
        # location)
        self.container_dir = self.host_dir

    def copy_input_files(self, logger):
        self._inputs_copied = True
        return super().copy_input_files(logger)

    def host_work_dir(self):
        # Since we aren't virtualizing the in-container paths as noted above, always use the same
        # working directory on task retries, instead of the base class behavior of appending the
        # try counter (on the host side). This loses some robustness to a split-brain condition
        # where the previous try is actually still running when we start the retry.
        # (see also retry_wait)
        return os.path.join(self.host_dir, "work")

    def host_stdout_txt(self):
        return os.path.join(self.host_dir, "stdout.txt")

    def host_stderr_txt(self):
        return os.path.join(self.host_dir, "stderr.txt")

    def reset(self, logger) -> None:
        cooldown = self.cfg.get_float("aws", "retry_wait")
        if cooldown > 0.0:
            logger.info(
                _(
                    "waiting to retry per configuration [aws] retry_wait",
                    seconds=cooldown,
                )
            )
            time.sleep(cooldown)

        rmtree_atomic(self.host_work_dir())
        with suppress(FileNotFoundError):
            os.unlink(self.host_stderr_txt() + ".offset")  # PygtailLogger state file
        super().reset(logger)

    def _run(self, logger, terminating, command):
        """
        Run task
        """
        self._observed_states = set()
        boto3_retries = self.cfg.get_dict("aws", "boto3_retries")
        try:
            aws_batch = boto3.Session().client(  # Session() needed for thread safety
                "batch",
                region_name=self._region_name,
                config=botocore.config.Config(retries=boto3_retries),
            )
            with ExitStack() as cleanup:
                # submit Batch job (with request throttling)
                job_id = None
                submit_period = self.cfg.get_float("aws", "submit_period")
                while True:
                    with self._submit_lock:
                        if terminating():
                            raise WDL.runtime.Terminated(quiet=True)
                        if (
                            time.time() - self._last_submit_time[0]
                            >= submit_period * self._submit_period_multiplier()
                        ):
                            job_id = self._submit_batch_job(logger, cleanup, aws_batch, command)
                            self._last_submit_time[0] = time.time()
                            break
                    time.sleep(submit_period / 4)
                # poll Batch job status
                return self._await_batch_job(logger, cleanup, aws_batch, job_id, terminating)
        except botocore.exceptions.ClientError as exn:
            wrapper = AWSError(exn)
            logger.error(wrapper)
            raise wrapper

    def _submit_batch_job(self, logger, cleanup, aws_batch, command):
        """
        Register & submit AWS batch job, leaving a cleanup callback to deregister the transient
        job definition.
        """

        job_name = self.run_id
        if job_name.startswith("call-"):
            job_name = job_name[5:]
        if self.try_counter > 1:
            job_name += f"-try{self.try_counter}"
        # Append entropy to the job name to avoid race condition using identical job names in
        # concurrent RegisterJobDefinition requests
        job_name = randomize_job_name(job_name)

        container_properties = self._prepare_container_properties(logger, command)
        job_def = aws_batch.register_job_definition(
            jobDefinitionName=job_name,
            type="container",
            containerProperties=container_properties,
        )
        job_def_handle = f"{job_def['jobDefinitionName']}:{job_def['revision']}"
        logger.debug(
            _(
                "registered Batch job definition",
                jobDefinition=job_def_handle,
                **container_properties,
            )
        )

        self._cleanup_job_definition(logger, cleanup, aws_batch, job_def_handle)

        job_tags = self.cfg.get_dict("aws", "job_tags")
        if "AWS_BATCH_JOB_ID" in os.environ:
            # If we find ourselves running inside an AWS Batch job, tag the new job identifying
            # ourself as the "parent" job.
            job_tags["AWS_BATCH_PARENT_JOB_ID"] = os.environ["AWS_BATCH_JOB_ID"]
        # TODO: set a tag to indicate that this job is a retry of another
        job = aws_batch.submit_job(
            jobName=job_name,
            jobQueue=self._job_queue,
            jobDefinition=job_def_handle,
            timeout={"attemptDurationSeconds": self.cfg.get_int("aws", "job_timeout")},
            tags=job_tags,
        )
        logger.info(
            _(
                "AWS Batch job submitted",
                jobQueue=self._job_queue,
                jobId=job["jobId"],
                tags=job_tags,
            )
        )
        return job["jobId"]

    def _prepare_container_properties(self, logger, command):
        image_tag = self.runtime_values.get("docker", "ubuntu:20.04")
        volumes, mount_points = self._prepare_mounts(logger, command)
        vcpu = self.runtime_values.get("cpu", 1)
        memory_mbytes = max(
            math.ceil(self.runtime_values.get("memory_reservation", 0) / 1048576), 1024
        )
        commands = [
            f"cd {self.container_dir}/work",
            "exit_code=0",
            "bash ../command >> ../stdout.txt 2> >(tee -a ../stderr.txt >&2) || exit_code=$?",
        ]
        if self.cfg.get_bool("aws", "container_sync"):
            commands.append("find . -type f | xargs sync")
            commands.append("sync ../stdout.txt ../stderr.txt")
        commands.append("exit $exit_code")

        container_properties = {
            "image": image_tag,
            "volumes": volumes,
            "mountPoints": mount_points,
            "command": ["/bin/bash", "-ec", "\n".join(commands)],
            "resourceRequirements": [
                {"type": "VCPU", "value": str(vcpu)},
                {"type": "MEMORY", "value": str(memory_mbytes)},
            ],
            "privileged": self.runtime_values.get("privileged", False),
        }

        if self.cfg["task_runtime"].get_bool("as_user"):
            user = (
                f"{self._studio_efs_uid}:{self._studio_efs_uid}"
                if self._studio_efs_uid is not None
                else f"{os.geteuid()}:{os.getegid()}"
            )
            if user.startswith("0:"):
                logger.warning(
                    "container command will run explicitly as root, since you are root and set --as-me"
                )
            container_properties["user"] = user

        return container_properties

    def _prepare_mounts(self, logger, command):
        """
        Prepare the "volumes" and "mountPoints" for the Batch job definition, assembling the
        in-container filesystem with the shared working directory, read-only input files, and
        command/stdout/stderr files.
        """

        # prepare control files
        with open(os.path.join(self.host_dir, "command"), "w") as outfile:
            outfile.write(command)
        with open(self.host_stdout_txt(), "w"):
            pass
        with open(self.host_stderr_txt(), "w"):
            pass

        # EFS mount point
        volumes = [
            {
                "name": "efs",
                "efsVolumeConfiguration": {
                    "fileSystemId": self._fs_id,
                    "transitEncryption": "ENABLED",
                },
            }
        ]
        if self._fsap_id:
            volumes[0]["efsVolumeConfiguration"]["authorizationConfig"] = {
                "accessPointId": self._fsap_id
            }
        mount_points = [{"containerPath": self._fs_mount, "sourceVolume": "efs"}]

        if self._inputs_copied:
            return volumes, mount_points

        # Prepare symlinks to the input Files & Directories
        container_prefix = os.path.join(self.container_dir, "work/_miniwdl_inputs/")
        link_dirs_made = set()
        for host_fn, container_fn in self.input_path_map.items():
            assert container_fn.startswith(container_prefix) and len(container_fn) > len(
                container_prefix
            )
            link_dn = os.path.dirname(container_fn)
            if link_dn not in link_dirs_made:
                os.makedirs(link_dn)
                link_dirs_made.add(link_dn)
            symlink_force(host_fn, container_fn)

        return volumes, mount_points

    def _cleanup_job_definition(self, logger, cleanup, aws_batch, job_def_handle):
        def deregister(logger, aws_batch, job_def_handle):
            try:
                aws_batch.deregister_job_definition(jobDefinition=job_def_handle)
                logger.debug(_("deregistered Batch job definition", jobDefinition=job_def_handle))
            except botocore.exceptions.ClientError as exn:
                # AWS expires job definitions after 6mo, so failing to delete them isn't fatal
                logger.warning(
                    _(
                        "failed to deregister Batch job definition",
                        jobDefinition=job_def_handle,
                        error=str(AWSError(exn)),
                    )
                )

        cleanup.callback(deregister, logger, aws_batch, job_def_handle)

    def _await_batch_job(self, logger, cleanup, aws_batch, job_id, terminating):
        """
        Poll for Batch job success or failure & return exit code
        """
        describe_period = self.cfg.get_float("aws", "describe_period")
        cleanup.callback((lambda job_id: self._describer.unsubscribe(job_id)), job_id)
        poll_stderr = cleanup.enter_context(
            PygtailLogger(logger, self.host_stderr_txt(), callback=self.stderr_callback)
        )
        exit_code = None
        while exit_code is None:
            time.sleep(describe_period)
            job_desc = self._describer.describe(aws_batch, job_id, describe_period)
            job_status = job_desc["status"]
            if "container" in job_desc and "logStreamName" in job_desc["container"]:
                self._logStreamName = job_desc["container"]["logStreamName"]
            if job_status not in self._observed_states:
                self._observed_states.add(job_status)
                logfn = (
                    logger.notice
                    if job_status in ("RUNNING", "SUCCEEDED", "FAILED")
                    else logger.info
                )
                logdetails = {"status": job_status, "jobId": job_id}
                if self._logStreamName:
                    logdetails["logStreamName"] = self._logStreamName
                logfn(_("AWS Batch job change", **logdetails))
                if job_status == "STARTING" or (
                    job_status == "RUNNING" and "STARTING" not in self._observed_states
                ):
                    # TODO: base TaskContainer should handle this, for separation of concerns
                    cleanup.enter_context(
                        WDL.runtime._statusbar.task_running(
                            self.runtime_values.get("cpu", 1),
                            self.runtime_values.get("memory_reservation", 0),
                        )
                    )
                if job_status not in (
                    "SUBMITTED",
                    "PENDING",
                    "RUNNABLE",
                    "STARTING",
                    "RUNNING",
                    "SUCCEEDED",
                    "FAILED",
                ):
                    logger.warning(_("unknown job status from AWS Batch", status=job_status))
            if job_status == "SUCCEEDED":
                exit_code = 0
            elif job_status == "FAILED":
                reason = job_desc.get("container", {}).get("reason", None)
                status_reason = job_desc.get("statusReason", None)
                self.failure_info = {"jobId": job_id}
                if reason:
                    self.failure_info["reason"] = reason
                if status_reason:
                    self.failure_info["statusReason"] = status_reason
                if self._logStreamName:
                    self.failure_info["logStreamName"] = self._logStreamName
                if status_reason and "Host EC2" in status_reason and "terminated" in status_reason:
                    raise WDL.runtime.Interrupted(
                        "AWS Batch job interrupted (likely spot instance termination)",
                        more_info=self.failure_info,
                    )
                if "exitCode" not in job_desc.get("container", {}):
                    raise WDL.Error.RuntimeError(
                        "AWS Batch job failed", more_info=self.failure_info
                    )
                exit_code = job_desc["container"]["exitCode"]
                assert isinstance(exit_code, int) and exit_code != 0
            if "RUNNING" in self._observed_states:
                poll_stderr()
            if terminating():
                aws_batch.terminate_job(jobId=job_id, reason="terminated by miniwdl")
                raise WDL.runtime.Terminated(
                    quiet=not self._observed_states.difference({"SUBMITTED", "PENDING", "RUNNABLE"})
                )
        for _root, _dirs, _files in os.walk(self.host_dir, followlinks=False):
            # no-op traversal of working directory to refresh NFS metadata cache (speculative)
            pass
        poll_stderr()
        return exit_code

    def _submit_period_multiplier(self):
        if self._describer.jobs:
            b = self.cfg.get_float("aws", "submit_period_b")
            if b > 0.0:
                t = time.time() - self._init_time
                c = self.cfg.get_float("aws", "submit_period_c")
                return max(1.0, c - t / b)
        return 1.0


class BatchJobDescriber:
    """
    This singleton object handles calling the AWS Batch DescribeJobs API with up to 100 job IDs
    per request, then dispensing each job description to the thread interested in it. This helps
    avoid AWS API request rate limits when we're tracking many concurrent jobs.
    """

    JOBS_PER_REQUEST = 100  # maximum jobs per DescribeJob request

    def __init__(self):
        self.lock = threading.Lock()
        self.last_request_time = 0
        self.job_queue = []
        self.jobs = {}

    def describe(self, aws_batch, job_id, period):
        """
        Get the latest Batch job description
        """
        while True:
            with self.lock:
                if job_id not in self.jobs:
                    # register new job to be described ASAP
                    heapq.heappush(self.job_queue, (0.0, job_id))
                    self.jobs[job_id] = None
                # update as many job descriptions as possible
                self._update(aws_batch, period)
                # return the desired job description if we have it
                desc = self.jobs[job_id]
                if desc:
                    return desc
            # otherwise wait (outside the lock) and try again
            time.sleep(period / 4)

    def unsubscribe(self, job_id):
        """
        Unsubscribe from a job_id once we'll no longer be interested in it
        """
        with self.lock:
            if job_id in self.jobs:
                del self.jobs[job_id]

    def _update(self, aws_batch, period):
        # if enough time has passed since our last DescribeJobs request
        if time.time() - self.last_request_time >= period:
            # take the N least-recently described jobs
            job_ids = set()
            assert self.job_queue
            while self.job_queue and len(job_ids) < self.JOBS_PER_REQUEST:
                job_id = heapq.heappop(self.job_queue)[1]
                assert job_id not in job_ids
                if job_id in self.jobs:
                    job_ids.add(job_id)
            if not job_ids:
                return
            # describe them
            try:
                job_descs = aws_batch.describe_jobs(jobs=list(job_ids))
            finally:
                # always: bump last_request_time and re-enqueue these jobs
                self.last_request_time = time.time()
                for job_id in job_ids:
                    heapq.heappush(self.job_queue, (self.last_request_time, job_id))
            # update self.jobs with the new descriptions
            for job_desc in job_descs["jobs"]:
                job_ids.remove(job_desc["jobId"])
                self.jobs[job_desc["jobId"]] = job_desc
            assert not job_ids, "AWS Batch DescribeJobs didn't return all expected results"


class AWSError(WDL.Error.RuntimeError):
    """
    Repackage botocore.exceptions.ClientError to surface it more-informatively in miniwdl task log
    """

    def __init__(self, client_error: botocore.exceptions.ClientError):
        assert isinstance(client_error, botocore.exceptions.ClientError)
        msg = (
            f"{client_error.response['Error']['Code']}, {client_error.response['Error']['Message']}"
        )
        super().__init__(
            msg, more_info={"ResponseMetadata": client_error.response["ResponseMetadata"]}
        )



"""
miniwdl_run_s3upload CLI entry point (console script) which passes through its arguments to
`miniwdl run`, then uploads run artifacts to $S3_UPLOAD_FOLDER. This includes the log file and if
the run succeeded, the output files and outputs.json (rewritten with the uploaded S3 URIs instead
of local filenames).

With the BatchJob plugin also enabled, this may be used from an SSH session on an EC2 instance or
container with EFS suitably mounted at /mnt/efs; or within a Batch "workflow job."
"""

import sys
import os
import json
import subprocess
import shutil
import argparse
import tempfile
import signal
from ._util import END_OF_LOG, subprocess_run_with_clean_exit


def miniwdl_run_s3upload():
    # Set signal handler. SystemExit may be handled below and/or by subprocess_run_with_clean_exit.
    for s in (signal.SIGTERM, signal.SIGINT, signal.SIGHUP, signal.SIGQUIT):
        signal.signal(s, lambda sig, _: (_ for _ in ()).throw(SystemExit(sig)))

    # run main logic with handlers
    try:
        end_log_and_exit(miniwdl_run_s3upload_inner())
    except SystemExit as exc:
        end_log_and_exit(exc.code)
    except KeyboardInterrupt:
        end_log_and_exit(int(signal.SIGINT))
    except BrokenPipeError:
        end_log_and_exit(int(signal.SIGPIPE))


def end_log_and_exit(code):
    print(
        "\n" + END_OF_LOG,
        file=sys.stderr,
    )
    sys.exit(code)


def miniwdl_run_s3upload_inner():
    parser = argparse.ArgumentParser(
        prog="miniwdl-run-s3upload",
        description="Pass through arguments to `miniwdl run` and afterwards, upload outputs to S3 and optionally delete local run directory.",
        usage="miniwdl-run-s3upload [miniwdl_run_arg ...]",
        allow_abbrev=False,
    )
    parser.add_argument(
        "--s3upload",
        help="s3://bucket/folder/ at which to upload run outputs [env MINIWDL__AWS__S3_UPLOAD_FOLDER]",
    )
    parser.add_argument(
        "--delete-after",
        choices=("always", "success", "failure"),
        help="with --s3upload, delete EFS run directory afterwards [env MINIWDL__AWS__S3_UPLOAD_DELETE_AFTER]",
    )
    parser.add_argument(
        "--task-queue", help="AWS Batch job queue for task jobs [env MINIWDL__AWS__TASK_QUEUE]"
    )

    args, unused_args = parser.parse_known_args(sys.argv[1:])
    args.s3upload = (
        args.s3upload if args.s3upload else os.environ.get("MINIWDL__AWS__S3_UPLOAD_FOLDER", None)
    )
    args.delete_after = (
        args.delete_after.strip().lower()
        if args.delete_after
        else os.environ.get("MINIWDL__AWS__S3_UPLOAD_DELETE_AFTER", None)
    )
    if args.delete_after and not args.s3upload:
        print("--delete-after requires --s3upload", file=sys.stderr)
        sys.exit(1)

    if args.s3upload:
        with tempfile.TemporaryDirectory() as tmpdir:
            testfile = os.path.join(tmpdir, ".test.miniwdl-run-s3upload")
            with open(testfile, "w") as outfile:
                print(
                    "miniwdl-run-s3upload created this object to test bucket permissions.",
                    file=outfile,
                )
            upload1(testfile, args.s3upload + ("/" if not args.s3upload.endswith("/") else ""))

    cmd = ["miniwdl", "run"] + unused_args
    if "--error-json" not in unused_args:
        cmd.append("--error-json")
    miniwdl_env = dict(os.environ)
    if args.task_queue:  # pass through to BatchJob plugin via env var
        miniwdl_env["MINIWDL__AWS__TASK_QUEUE"] = args.task_queue

    # run miniwdl & tee its standard output
    miniwdl = subprocess_run_with_clean_exit(
        cmd, stdout=subprocess.PIPE, env=miniwdl_env, check=False
    )
    sys.stdout.buffer.write(miniwdl.stdout)

    if not args.s3upload:
        # nothing to do
        print(
            f"[miniwdl_run_s3upload] no setting for --s3upload / MINIWDL__AWS__S3_UPLOAD_FOLDER; exiting (code = {miniwdl.returncode})",
            file=sys.stderr,
        )
        return miniwdl.returncode

    # read miniwdl standard output JSON
    try:
        miniwdl_json = json.loads(miniwdl.stdout)
        run_dir = miniwdl_json["dir"]
        assert os.path.isdir(run_dir)
    except:
        print(
            f"[miniwdl_run_s3upload] no run directory in miniwdl standard output; exiting (code = {miniwdl.returncode})",
            file=sys.stderr,
        )
        return miniwdl.returncode

    # append miniwdl's run name to S3_UPLOAD_FOLDER (unless the latter ends in '/')
    s3_upload_folder = args.s3upload
    if not s3_upload_folder.endswith("/"):
        s3_upload_folder += os.path.basename(run_dir.rstrip("/")) + "/"

    # upload logs
    print(
        f"[miniwdl_run_s3upload] miniwdl exit code = {miniwdl.returncode}; uploading logs & outputs to {s3_upload_folder}",
        file=sys.stderr,
    )
    for p in (os.path.join(run_dir, fn) for fn in ("workflow.log", "task.log")):
        if os.path.isfile(p):
            upload1(p, s3_upload_folder)

    # upload error.json, and the stderr_file it points to, if any
    error_json_file = os.path.join(run_dir, "error.json")
    if os.path.isfile(error_json_file):
        upload1(error_json_file, s3_upload_folder)
        with open(error_json_file) as infile:
            error_json = json.load(infile)
            stderr_file = error_json.get("cause", {}).get("stderr_file", None)
            if stderr_file and os.path.isfile(stderr_file):
                stderr_s3file = s3_upload_folder + "CommandFailed_stderr.txt"
                upload1(stderr_file, stderr_s3file)
                error_json["cause"]["stderr_s3file"] = stderr_s3file
                with tempfile.NamedTemporaryFile() as tmp:
                    tmp.write(json.dumps(error_json, indent=2).encode())
                    tmp.flush()
                    upload1(tmp.name, s3_upload_folder + "error.json")

    # upload output files, if any
    if os.path.isdir(os.path.join(run_dir, "out")):
        subprocess_run_with_clean_exit(
            [
                "aws",
                "s3",
                "sync",
                "--no-progress",
                "--follow-symlinks",
                os.path.join(run_dir, "out"),
                s3_upload_folder,
            ],
            check=True,
        )

    if "outputs" not in miniwdl_json:
        if args.delete_after in ("always", "failure"):
            shutil.rmtree(run_dir)
            print(
                f"[miniwdl_run_s3upload] deleted {run_dir}",
                file=sys.stderr,
            )
        return miniwdl.returncode

    # recursively rewrite outputs JSON
    def rewrite(v):

        if v and isinstance(v, str) and v[0] == "/" and os.path.exists(v):
            # miniwdl writes File/Directory outputs with absolute paths
            return rebase_output_path(v, run_dir, s3_upload_folder)
        if isinstance(v, list):
            return [rewrite(u) for u in v]
        if isinstance(v, dict):
            return dict((k, rewrite(u)) for (k, u) in v.items())
        return v

    rewritten_outputs = rewrite(miniwdl_json["outputs"])
    outputs_s3_json = os.path.join(run_dir, "outputs.s3.json")
    with open(outputs_s3_json + ".tmp", "w") as outfile:
        print(json.dumps(rewritten_outputs, indent=2), file=outfile)
    os.rename(outputs_s3_json + ".tmp", outputs_s3_json)
    upload1(outputs_s3_json, s3_upload_folder + "outputs.json")
    print(
        f"[miniwdl_run_s3upload] uploaded {s3_upload_folder}outputs.json",
        file=sys.stderr,
    )
    print(json.dumps({"s3upload": s3_upload_folder, "outputs": rewritten_outputs}, indent=2))
    if args.delete_after in ("always", "success"):
        shutil.rmtree(run_dir)
        print(
            f"[miniwdl_run_s3upload] deleted {run_dir}",
            file=sys.stderr,
        )

    return miniwdl.returncode


def upload1(fn, dest):
    subprocess_run_with_clean_exit(["aws", "s3", "cp", "--no-progress", fn, dest], check=True)


def rebase_output_path(fn, run_dir, s3_upload_folder):
    """
    Given extant filename `fn` from JSON outputs and the current run directory, figure the uploaded
    S3 URI under s3_upload_folder, where the file should be uploaded by our `aws s3 sync` operation
    on the "run out" directory. Or return fn unmodified if it seems to be something that looks like
    an output path, but isn't really.

    Subtlety: if the output fn originated from the call cache, it will be from some other run
    directory, not the current one. In that case we need to see that there's a corresponding link
    under the current run out directory.

    There should be no danger of inadvertently uploading non-output files (e.g. if the workflow
    outputs the string "/home/root/.ssh/id_rsa") because we're not actually performing the upload,
    just figuring the path where `aws s3 sync` ought to have uploaded it.
    """
    fn_parts = fn.strip("/").split("/")
    while fn_parts:
        fn_rel = "/".join(fn_parts)
        fn_rebased = os.path.join(run_dir, "out", fn_rel)
        if os.path.exists(fn_rebased) and os.path.isdir(fn) == os.path.isdir(fn_rebased):
            return s3_upload_folder + fn_rel
        fn_parts = fn_parts[1:]
    return fn
